from dependency_paraphraser.pretrained_projectors import natasha_projector


from natasha import (
    Segmenter,
    MorphVocab,
    NewsEmbedding,
    NewsMorphTagger,
    NewsSyntaxParser,
    Doc,
)

from dependency_paraphraser import projection, synonyms


segmenter = Segmenter()
morph_vocab = MorphVocab()

emb = NewsEmbedding()
morph_tagger = NewsMorphTagger(emb)
syntax_parser = NewsSyntaxParser(emb)

gensim_emb = None


def use_news_embeddings():
    """ Convert navec embeddings to gensim format to use for synonym replacement """
    global gensim_emb
    gensim_emb = emb.as_gensim


def paraphrase(text, tree_temperature=0.5, w2v=None, min_sim=0.5, p_rep=0.5, projector=natasha_projector):
    doc = Doc(text)
    doc.segment(segmenter)
    doc.tag_morph(morph_tagger)
    doc.parse_syntax(syntax_parser)

    if w2v is None:
        w2v = gensim_emb

    results = []

    for sent in doc.sents:
        toks = projection.make_tree_projection(
            sent, model=projector, temperature=tree_temperature,
        )
        if w2v:
            words = synonyms.replace_synonyms(
                toks, w2v=w2v, morph_vocab=morph_vocab, min_sim=min_sim, p_rep=p_rep,
            )
        else:
            words = [token.text for token in toks]
        results.append(' '.join(words))
    return ' '.join(results)



import random
from functools import cmp_to_key


from dependency_paraphraser.utils import anneal, nat2ch, conll2ch


def sent2xy(sent, sentence_format='natasha'):
    pairs_x = []
    pairs_y = []
    pairs_ids = []
    tree = nat2ch(sent) if sentence_format == 'natasha' else conll2ch(sent)
    tokens = sent.tokens if sentence_format == 'natasha' else sent

    def tok2rel(tok):
        if sentence_format == 'natasha':
            return tok.rel
        return tok.deprel

    def tok2pos(tok):
        if sentence_format == 'natasha':
            return tok.pos
        return tok.upos

    def tok2features(tok, idx):
        dep = 'parent' if idx == parent_id else tok2rel(tok)
        return {
            'deprel_{}'.format(dep): 1,
            'pos_{}'.format(tok2pos(tok)): 1,
        }

    for parent_id, children_ids in tree.items():
        ids = children_ids if parent_id == -1 else [parent_id] + children_ids
        for l_id in ids:
            for r_id in ids:
                if l_id == r_id:
                    continue
                pairs_x.append({
                    '{}_{}'.format(lr, k): v
                    for lr, idx in [('r', r_id), ('l', l_id)]
                    for k, v in tok2features(tokens[idx], idx).items()
                })
                pairs_y.append(int(l_id < r_id))
                pairs_ids.append((l_id, r_id))
    return pairs_x, pairs_y, pairs_ids


def find_roots(id2children):
    """ Find the set of roots - the top nodes of all trees in the graph """
    id2parent = {}
    id2root = {}

    for parent_id, children_ids in id2children.items():
        for child_id in children_ids:
            id2parent[child_id] = parent_id

    def find_root(node_id, stack=None):
        parent_id = id2parent.get(node_id)
        if parent_id in id2root:
            return id2root[parent_id]
        if parent_id == node_id or parent_id is None:
            root_id = node_id
        elif parent_id == -1:
            root_id = parent_id
        else:
            if stack is None:
                stack = set()
            if node_id in stack:  # the "tree" is cyclic
                root_id = node_id
            else:
                stack.add(node_id)
                root_id = find_root(parent_id, stack=stack)
        id2root[node_id] = root_id
        return root_id

    for child_id in id2parent:
        find_root(child_id)

    roots = set(id2root.values())
    return roots


def project_tree_randomly(id_to_children, pair_to_proba=None, root_id=-1, temperature=1, shuffle=True):
    """
    Project a tree into a sequence, optionally with a random order of children.
    Return the list of token indices.
    """
    if root_id not in id_to_children:
        if root_id == -1:  # a sentence with empty root and no children is empty
            return []
        return [root_id]
    children_ids = [c for c in id_to_children[root_id] if c != root_id]

    ids = children_ids if root_id == -1 else [root_id] + children_ids

    def comparator(l, r):
        raw_proba = (pair_to_proba[(l, r)] + 1 - pair_to_proba[(r, l)]) / 2
        new_proba = anneal(raw_proba, t=temperature)
        return random.random() - new_proba

    if pair_to_proba:
        random.shuffle(ids)
        ids = sorted(ids, key=cmp_to_key(comparator))
    elif shuffle:
        random.shuffle(ids)
    else:
        ids = sorted(ids)

    result = []
    for tok_id in ids:
        if tok_id == root_id:
            result.append(tok_id)
        else:
            result.extend(project_tree_randomly(
                id_to_children,
                root_id=tok_id,
                pair_to_proba=pair_to_proba,
                temperature=temperature,
                shuffle=shuffle,
            ))
    return result


def make_tree_projection(sent, model, sentence_format='natasha', temperature=1):
    px, py, pids = sent2xy(sent, sentence_format=sentence_format)
    preds = model.predict_proba(px)[:, 1]
    pair2proba = {pair: proba for pair, proba in zip(pids, preds)}
    ch = nat2ch(sent) if sentence_format == 'natasha' else conll2ch(sent)
    tokens = sent.tokens if sentence_format == 'natasha' else sent
    roots = find_roots(ch)
    # filter out possible recursion
    ch = {k: [v for v in vv if v not in roots] for k, vv in ch.items()}
    positions = [
        t
        for root_id in roots
        for t in project_tree_randomly(ch, pair_to_proba=pair2proba, temperature=temperature, root_id=root_id)
    ]
    return [tokens[t] for t in positions]



import ufal.udpipe
import pyconll

from dependency_paraphraser import synonyms, projection
from dependency_paraphraser.pretrained_projectors import natasha_projector, en_udpipe_projector


class Model:
    def __init__(self, path):
        """Load given model."""
        self.model = ufal.udpipe.Model.load(path)
        if not self.model:
            raise Exception("Cannot load UDPipe model from file '%s'" % path)

    def tokenize(self, text):
        """Tokenize the text and return list of ufal.udpipe.Sentence-s."""
        tokenizer = self.model.newTokenizer(self.model.DEFAULT)
        if not tokenizer:
            raise Exception("The model does not have a tokenizer")
        return self._read(text, tokenizer)

    def read(self, text, in_format):
        """Load text in the given format (conllu|horizontal|vertical) and return list of ufal.udpipe.Sentence-s."""
        input_format = ufal.udpipe.InputFormat.newInputFormat(in_format)
        if not input_format:
            raise Exception("Cannot create input format '%s'" % in_format)
        return self._read(text, input_format)

    def _read(self, text, input_format):
        input_format.setText(text)
        error = ufal.udpipe.ProcessingError()
        sentences = []

        sentence = ufal.udpipe.Sentence()
        while input_format.nextSentence(sentence, error):
            sentences.append(sentence)
            sentence = ufal.udpipe.Sentence()
        if error.occurred():
            raise Exception(error.message)

        return sentences

    def tag(self, sentence):
        """Tag the given ufal.udpipe.Sentence (inplace)."""
        self.model.tag(sentence, self.model.DEFAULT)

    def parse(self, sentence):
        """Parse the given ufal.udpipe.Sentence (inplace)."""
        self.model.parse(sentence, self.model.DEFAULT)

    def write(self, sentences, out_format):
        """Write given ufal.udpipe.Sentence-s in the required format (conllu|horizontal|vertical)."""

        output_format = ufal.udpipe.OutputFormat.newOutputFormat(out_format)
        output = ''
        for sentence in sentences:
            output += output_format.writeSentence(sentence)
        output += output_format.finishDocument()

        return output

    def __call__(self, text):
        sentences = self.tokenize(text)
        for s in sentences:
            self.tag(s)
            self.parse(s)
        conllu = self.write(sentences, "conllu")
        result = pyconll.load_from_string(conllu)
        return result


def paraphrase(
        text,
        udpipe_model,
        tree_temperature=0.5,
        w2v=None,
        min_sim=0.5,
        p_rep=0.5,
        projector=natasha_projector,
        morph=None,
):
    doc = udpipe_model(text)
    results = []
    for sent in doc:
        toks = projection.make_tree_projection(
            sent, model=projector, temperature=tree_temperature, sentence_format='conll',
        )
        if w2v:
            words = synonyms.replace_synonyms(
                toks, w2v=w2v, morph_vocab=morph, min_sim=min_sim, p_rep=p_rep,
            )
        else:
            words = [token.form for token in toks]
        results.append(' '.join(words))
    return ' '.join(results)



import random


INFLECTABLES = {
    'NOUN': ['Number', 'Case'],
    'ADJ': ['Number', 'Case', 'Gender',], #  'Degree' ??
    'VERB': ['Number', 'VerbForm', 'Person', 'Tense', 'Mood', 'Gender'],
}
DEFAULT_INFLECTABLES = ['Number', 'Case', 'VerbForm', 'Person', 'Tense', 'Mood']  # 'Gender',


def morph_synonyms(token, w2v, morph_vocab=None, initial_k=30, k=10, threshold=0.0):
    if morph_vocab and hasattr(token, 'lemmatize'):
        # natasha tokens require lemmatization
        token.lemmatize(morph_vocab)
    text = token.lemma
    neighbours = []
    if text not in w2v:
        return neighbours
    pairs = w2v.most_similar(text, topn=initial_k)
    if not morph_vocab:
        return [pair for pair in pairs[:k] if pair[-1] >= threshold]
    for pair in pairs:
        if len(neighbours) >= k:
            break
        if pair[1] < threshold:
            break
        parses = morph_vocab.parse(pair[0])
        for parse in parses:
            if parse.normal == text:
                continue
            if parse.pos == token.pos:
                inflectables = INFLECTABLES.get(token.pos, DEFAULT_INFLECTABLES)
                word = parse.inflect({token.feats[k] for k in inflectables if k in token.feats})
                if not word:
                    continue
                if word.word == text:
                    break
                neighbours.append((word.word, pair[1]))
                break
    return neighbours


def replace_synonyms(tokens, w2v, morph_vocab=None, min_sim=0.6, p_rep=0.5):
    result = []
    for token in tokens:
        if hasattr(token, 'text'):
            text = token.text
        elif hasattr(token, 'form'):
            text = token.form
        else:
            text = str(token)
        if random.random() > p_rep:
            result.append(text)
            continue
        neighbors = morph_synonyms(token, w2v, morph_vocab=morph_vocab, threshold=min_sim)

        if neighbors:
            result.append(random.choice(neighbors)[0])
        else:
            result.append(text)
    return result



import math

from collections import defaultdict


def conll2ch(sentence):
    id2children = defaultdict(list)
    for tok in sentence:
        id2children[int(tok.head) - 1].append(int(tok.id)-1)
    return id2children


def nat2ch(sentence):
    id2idx = {token.id: i for i, token in enumerate(sentence.tokens)}
    id2children = defaultdict(list)
    for tok in sentence.tokens:
        id2children[id2idx.get(tok.head_id, -1)].append(id2idx[tok.id])
    return id2children


def anneal(p, t=1):
    """ Modify a probability by applying temperature to it. """
    if t == 1:
        return p
    if t == math.inf:
        return 0.5
    if t == 0:
        return int(p > 0.5)
    return 1 / (1 + math.exp(-(math.log(p / (1-p)) / t)))



import pickle

from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import make_pipeline
from sklearn.feature_extraction import DictVectorizer
from sklearn.model_selection import cross_val_score

from dependency_paraphraser.projection import sent2xy


def label_natasha_sentences(texts):
    from dependency_paraphraser.natasha import Doc, segmenter, morph_tagger, syntax_parser
    sentences = []
    for text in texts:
        doc = Doc(text)
        doc.segment(segmenter)
        doc.tag_morph(morph_tagger)
        doc.parse_syntax(syntax_parser)
        sentences.extend(doc.sents)
    return sentences


def label_udpipe_sentences(texts, model):
    sentences = []
    for text in texts:
        doc = model(text)
        sentences.extend(list(doc))
    return sentences


def load_conll_sentences(filenames):
    import pyconll
    sentences = []
    for fn in filenames:
        sentences.extend(list(
            pyconll.load_from_file(fn)
        ))
    return sentences


def train_projector(parsed_sentences, sentence_format='conll', output_file=None, cv=None):
    all_px = []
    all_py = []

    for sent in parsed_sentences:
        px, py, pids = sent2xy(sent, sentence_format=sentence_format)
        all_px.extend(px)
        all_py.extend(py)

    model = make_pipeline(
        DictVectorizer(sparse=True, dtype=int),
        PolynomialFeatures(include_bias=False, interaction_only=True),
        LogisticRegression(C=1, max_iter=1000),
    )
    if cv:
        print('Cross-validated ROC AUC scores:')
        print(cross_val_score(model, all_px, all_py, cv=cv, scoring='roc_auc'))
    model.fit(all_px, all_py)

    if output_file:
        with open('natasha_projector.pkl', 'wb') as f:
            pickle.dump(model, f)

    return model



import os
import pickle
import sklearn  # noqa

models_path = os.path.join(os.path.dirname(__file__), 'models')

with open(os.path.join(models_path, 'natasha_projector.pkl'), 'rb') as f:
    natasha_projector = pickle.load(f)

with open(os.path.join(models_path, 'en_udpipe_projector.pkl'), 'rb') as f:
    en_udpipe_projector = pickle.load(f)
